{
 "metadata": {
  "name": "",
  "signature": "sha256:882ada8905b3502c42566bc1903e872f92c0b9ee543c814d8fd09cf9a73fdfd0"
 },
 "nbformat": 3,
 "nbformat_minor": 0,
 "worksheets": [
  {
   "cells": [
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "%matplotlib inline\n",
      "import numpy as np\n",
      "import scipy.sparse.linalg as la"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 1
    },
    {
     "cell_type": "heading",
     "level": 1,
     "metadata": {},
     "source": [
      "Heisenberg Model"
     ]
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "Cout how many bits are set:"
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "POPCOUNT8 = [0] * 2**8\n",
      "for index in xrange(len(POPCOUNT8)):\n",
      "    POPCOUNT8[index] = (index & 1) + POPCOUNT8[index >> 1]\n",
      "\n",
      "def popcount32(x):\n",
      "    return POPCOUNT8[x & 0xff] + POPCOUNT8[(x >> 8) & 0xff] + POPCOUNT8[(x >> 16) & 0xff] + POPCOUNT8[(x >> 24) & 0xff]"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 2
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "Spin-$1/2$ basis for given number of up-spins."
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "INDEX_TYPE = np.int64\n",
      "INVALID_INDEX = -1\n",
      "\n",
      "class SpinOneHalfBasis:\n",
      "    def __init__(self, num_spins, num_ups):\n",
      "        assert 0 <= num_ups <= num_spins\n",
      "        assert num_spins <= 32\n",
      "        \n",
      "        self.num_spins = num_spins\n",
      "        self.num_ups = num_ups\n",
      "        \n",
      "        # maps state to its index in the state array\n",
      "        self.states = []        \n",
      "        self.index = INVALID_INDEX * np.ones(2**num_spins, dtype=INDEX_TYPE)\n",
      "        for s in xrange(2**num_spins):\n",
      "            if popcount32(s) == num_ups:\n",
      "                self.index[s] = len(self.states)\n",
      "                self.states.append(s)\n",
      "        self.dim = len(self.states)                "
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 3
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "Heisenberg Hamiltonian:"
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "def heisenberg_hamiltonian(num_spins, num_ups, J, periodic=False):\n",
      "    basis = SpinOneHalfBasis(num_spins, num_ups)\n",
      "    num_pairs = num_spins - 1\n",
      "    D = basis.dim\n",
      "    \n",
      "    def matvec(v):\n",
      "        w = np.empty(D)\n",
      "        \n",
      "        # \\sum_j sigma_z^j sigma_z^(j+1)\n",
      "        mask = 2**num_pairs - 1\n",
      "        for idx, s in enumerate(basis.states):\n",
      "            # count number of parallel & antiparallel pairs of spins\n",
      "            p = popcount32(mask & ~(s ^ (s>>1)))\n",
      "            a = num_pairs - p\n",
      "            \n",
      "            # set diagonal entry\n",
      "            w[idx] = 0.25 * (p - a) * v[idx]\n",
      "        \n",
      "        # \\sum_j sigma_+^j sigma_-^(j+1) + h.c.\n",
      "        for idx, s in enumerate(basis.states):\n",
      "            for r in xrange(num_pairs):\n",
      "                # flip [01, 10] -> 0.5 * [10, 01]\n",
      "                idx2 = basis.index[s ^ (0b11 << r)]  # will be INVALID_INDEX if we had [00, 11] at this position, since then the number of up-spins is changed\n",
      "                if idx2 != INVALID_INDEX:\n",
      "                    w[idx2] += 0.5 * v[idx]\n",
      "        \n",
      "        # sigma_+^1 sigma_-^L + h.c.\n",
      "        if periodic:\n",
      "            for idx, s in enumerate(basis.states):\n",
      "                # diagonal\n",
      "                p = 1 & ~(s ^ (s>>num_pairs))\n",
      "                a = 1 - p\n",
      "                w[idx] += 0.25 * (p - a) * v[idx]\n",
      "                \n",
      "                # offdiagonal\n",
      "                idx2 = basis.index[s ^ (1 << num_pairs | 1)]\n",
      "                if idx2 != INVALID_INDEX:\n",
      "                    w[idx2] += 0.5 * v[idx]\n",
      "\n",
      "        return J * w\n",
      "\n",
      "    return la.LinearOperator((D, D), matvec, dtype=np.float64)\n",
      "\n",
      "assert np.isclose(heisenberg_hamiltonian(2, 2, J=1) * [1], [0.25])\n",
      "assert np.isclose(heisenberg_hamiltonian(2, 0, J=1) * [1], [0.25])\n",
      "assert np.allclose(heisenberg_hamiltonian(2, 1, J=1) * [1, 0], [-0.25, 0.5])\n",
      "assert np.allclose(heisenberg_hamiltonian(2, 1, J=1) * [0, 1], [0.5, -0.25])"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 4
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "Compute energies of lowest eigenstates:"
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "def heisenberg_energies(k, num_spins, num_ups, J, periodic):\n",
      "    H = heisenberg_hamiltonian(num_spins, num_ups, J, periodic)\n",
      "    \n",
      "    # 1x1 Hamiltonian\n",
      "    if H.shape == (1, 1):\n",
      "        return H * [1] / num_spins\n",
      "\n",
      "    # compute using lanczos\n",
      "    Es = la.eigsh(H, k, which='SA', return_eigenvectors=False)\n",
      "    return Es / num_spins\n",
      "\n",
      "heisenberg_energies(1, num_spins=2, num_ups=1, J=1, periodic=False)"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [
      {
       "metadata": {},
       "output_type": "pyout",
       "prompt_number": 5,
       "text": [
        "array([-0.375])"
       ]
      }
     ],
     "prompt_number": 5
    },
    {
     "cell_type": "heading",
     "level": 1,
     "metadata": {},
     "source": [
      "Test Drive..."
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "%%time\n",
      "for L in range(2, 10):\n",
      "    print L, heisenberg_energies(1, L, L/2, 1, periodic=False)"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [
      {
       "output_type": "stream",
       "stream": "stdout",
       "text": [
        "2 [-0.375]\n",
        "3 [-0.33333333]\n",
        "4 [-0.40400635]\n",
        "5 [-0.38557725]\n",
        "6 [-0.41559619]\n",
        "7 "
       ]
      },
      {
       "output_type": "stream",
       "stream": "stdout",
       "text": [
        "[-0.4051771]\n",
        "8 "
       ]
      },
      {
       "output_type": "stream",
       "stream": "stdout",
       "text": [
        "[-0.42186657]\n",
        "9 "
       ]
      },
      {
       "output_type": "stream",
       "stream": "stdout",
       "text": [
        "[-0.41514686]\n",
        "Wall time: 505 ms\n"
       ]
      }
     ],
     "prompt_number": 6
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "%%time\n",
      "for L in range(2, 10):\n",
      "    print L, heisenberg_energies(1, L, L/2, 1, periodic=True)"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [
      {
       "output_type": "stream",
       "stream": "stdout",
       "text": [
        "2 [-0.75]\n",
        "3 [-0.25]\n",
        "4 [-0.5]\n",
        "5 [-0.3736068]\n",
        "6 [-0.46712927]\n",
        "7 "
       ]
      },
      {
       "output_type": "stream",
       "stream": "stdout",
       "text": [
        "[-0.40788275]\n",
        "8 "
       ]
      },
      {
       "output_type": "stream",
       "stream": "stdout",
       "text": [
        "[-0.45638668]\n",
        "9 "
       ]
      },
      {
       "output_type": "stream",
       "stream": "stdout",
       "text": [
        "[-0.4219222]\n",
        "Wall time: 424 ms\n"
       ]
      }
     ],
     "prompt_number": 7
    }
   ],
   "metadata": {}
  }
 ]
}